"""This module contains the exception thrower."""
from __future__ import annotations


def _raise_exception(exception: type, exception_message: str | None):
    """This function checks if a exception message is given.

    If there is no exception message, the default behavior is
    maintained. If there is an exception message, the message is passed
    to the exception with the 'value' keyword.

    :param exception_type: exception to raise when the timeout
        is reached.
    :type exception_type: type
    :param exception_message: optional message to pass to the exception
        when the timeout is reached.

    :raises: exception of type exception_type.
    """
    if exception_message is None:
        raise exception()
    raise exception(exception_message)



"""This module contains the wrappers for the timeout decorator."""
from __future__ import annotations

import signal
from functools import wraps
from typing import Callable

from timeoutd._timeout import _Timeout
from timeoutd.exceptions import _raise_exception


def _signaler(
    function: Callable,
    seconds: float | None,
    use_signals: bool,
    exception_type: type,
    exception_message: str | None,
) -> Callable:
    """This function checks if signals should be used for timing out.

    If the use_signals flag is set to True, the signaler is used. If the
    flag is set to False, the multiprocessing is used.

    :param function: function to wrap
    :type function: Callable
    :param seconds: optional time limit in seconds or fractions of a
        second. If None is passed, no timeout is applied.
        This adds some flexibility to the usage: you can disable timing
        out depending on the settings.
    :type seconds: float
    :param use_signals: flag indicating whether signals should be used
        for timing function out or the multiprocessing.
        When using multiprocessing, timeout granularity is limited to
        10ths of a second.
    :type use_signals: bool
    :param exception_type: exception to raise when the timeout is
        reached.
    :type exception_type: type
    :param exception_message: optional message to pass to the exception
        when the timeout is reached.

    :return: wrapped function

    :raises: TimeoutError if time limit is reached.
    """
    if use_signals:

        def handler(*args, **kwargs):  # pylint: disable=unused-argument
            _raise_exception(exception_type, exception_message)

        @wraps(function)
        def new_function(*args, **kwargs):
            new_seconds = kwargs.pop("timeout", seconds)
            if new_seconds:
                old_handler = signal.signal(signal.SIGALRM, handler)
                signal.setitimer(signal.ITIMER_REAL, new_seconds)

            if not seconds:
                return function(*args, **kwargs)

            try:
                return function(*args, **kwargs)
            finally:
                if new_seconds:
                    signal.setitimer(signal.ITIMER_REAL, 0)
                    # reinstall the old signal handler
                    signal.signal(signal.SIGALRM, old_handler)

        return new_function

    @wraps(function)
    def new_mt_function(*args, **kwargs):
        timeout_wrapper = _Timeout(
            function=function,
            exception_type=exception_type,
            exception_message=exception_message,
            limit=seconds,
        )
        return timeout_wrapper(*args, **kwargs)

    return new_mt_function


def _exception_handler(
    function: Callable,
    on_timeout: Callable,
    *,
    exception_type: type,
    on_timeout_args: tuple | None,
    on_timeout_kwargs: dict | None,
) -> Callable:
    """This function catches the exception and calls the on_timeout function.

    :param function: function to wrap
    :type function: Callable
    :type on_timeout: Callable
    :param exception_type: optional exception to raise when the timeout
        is reached. If None is passed, the default behavior is to raise
        a TimeoutError exception.
    :type exception_type: type
    :param on_timeout_args: optional arguments to pass to the on_timeout
        function.
    :type on_timeout_args: tuple
    :param on_timeout_kwargs: optional keyword arguments to pass to the
        on_timeout function.
    :type on_timeout_kwargs: dict

    :return: wrapped function
    """
    if on_timeout_args is None:
        on_timeout_args = ()
    if on_timeout_kwargs is None:
        on_timeout_kwargs = {}

    @wraps(function)
    def new_function(*args, **kwargs):
        try:
            return function(*args, **kwargs)
        except exception_type:
            return on_timeout(*on_timeout_args, **on_timeout_kwargs)

    return new_function



"""Timeout decorator tests."""
import time

import pytest

from timeoutd import timeout

TIMEOUT = 0.1
EXCEPTION_MESSAGE = "Timeout exceeded."


@pytest.fixture(params=[False, True])
def use_signals(request):
    """Use signals for timing out or not."""
    return request.param


def test_timeout_decorator_arg(use_signals):
    @timeout(seconds=TIMEOUT, use_signals=use_signals)
    def f():
        time.sleep(2)

    with pytest.raises(TimeoutError):
        f()


def test_timeout_class_method(use_signals):
    class c:
        @timeout(seconds=TIMEOUT, use_signals=use_signals)
        def f(self):
            time.sleep(2)

    with pytest.raises(TimeoutError):
        c().f()


def test_timeout_kwargs(use_signals):
    @timeout(seconds=0.3, use_signals=use_signals)
    def f():
        time.sleep(0.2)

    with pytest.raises(TimeoutError):
        f(timeout=0.1)


def test_timeout_alternate_exception(use_signals):
    @timeout(seconds=0.3, use_signals=use_signals, exception_type=StopIteration)
    def f():
        time.sleep(0.2)

    with pytest.raises(StopIteration):
        f(timeout=0.1)


def test_timeout_kwargs_with_initial_timeout_none(use_signals):
    @timeout(use_signals=use_signals)
    def f():
        time.sleep(0.2)

    with pytest.raises(TimeoutError):
        f(timeout=0.1)


def test_timeout_no_seconds(use_signals):
    @timeout(use_signals=use_signals)
    def f():
        time.sleep(0.1)

    f()


def test_timeout_partial_seconds(use_signals):
    @timeout(seconds=0.2, use_signals=use_signals)
    def f():
        time.sleep(0.5)

    with pytest.raises(TimeoutError):
        f()


def test_timeout_ok(use_signals):
    @timeout(seconds=0.2, use_signals=use_signals)
    def f():
        time.sleep(0.1)

    f()


def test_function_name(use_signals):
    @timeout(seconds=0.2, use_signals=use_signals)
    def func_name():
        pass

    assert func_name.__name__ == "func_name"


def test_timeout_custom_on_timeout(use_signals):
    def on_timeout():
        return 0

    @timeout(seconds=TIMEOUT, on_timeout=on_timeout, use_signals=use_signals)
    def f():
        time.sleep(2)

    assert f() == 0


def test_timeout_custom_on_timeout_with_args(use_signals):
    def on_timeout(i, j):
        return i + j

    @timeout(
        seconds=TIMEOUT,
        on_timeout=on_timeout,
        use_signals=use_signals,
        on_timeout_args=(1, 2),
    )
    def f():
        time.sleep(2)

    assert f() == 3


def test_timeout_custom_on_timeout_with_kwargs(use_signals):
    def on_timeout(i=0, j=1):
        return i + j

    @timeout(
        seconds=TIMEOUT,
        on_timeout=on_timeout,
        use_signals=use_signals,
        on_timeout_kwargs={"i": 1, "j": 2},
    )
    def f():
        time.sleep(2)

    assert f() == 3


def test_timeout_custom_on_timeout_with_args_and_kwargs(use_signals):
    def on_timeout(i, j=1):
        return i + j

    @timeout(
        seconds=TIMEOUT,
        on_timeout=on_timeout,
        use_signals=use_signals,
        on_timeout_args=(1,),
        on_timeout_kwargs={"j": 2},
    )
    def f():
        time.sleep(2)

    assert f() == 3


def test_timeout_pickle_error():
    """Test that when a pickle error occurs a timeout error is raised."""

    @timeout(seconds=TIMEOUT, use_signals=False)
    def f():
        time.sleep(0.1)

        class Test:
            pass

        return Test()

    with pytest.raises(TimeoutError):
        f()


def test_timeout_custom_exception_message():
    @timeout(seconds=TIMEOUT, exception_message=EXCEPTION_MESSAGE)
    def f():
        time.sleep(2)

    with pytest.raises(TimeoutError, match=EXCEPTION_MESSAGE):
        f()


def test_timeout_custom_exception_with_message():
    @timeout(
        seconds=TIMEOUT,
        exception_type=RuntimeError,
        exception_message=EXCEPTION_MESSAGE,
    )
    def f():
        time.sleep(2)

    with pytest.raises(RuntimeError, match=EXCEPTION_MESSAGE):
        f()


def test_timeout_exception_is_of_wrong_type():
    with pytest.raises(TypeError):

        @timeout(seconds=TIMEOUT, exception_type=str)
        def f():
            time.sleep(2)



"""Timeout decorator."""

from __future__ import annotations

from typing import Callable

from timeoutd.wrappers import _exception_handler, _signaler


def timeout(
    seconds: float | None = None,
    on_timeout: Callable | None = None,
    *,
    use_signals: bool = True,
    exception_type: type = TimeoutError,
    exception_message: str | None = None,
    on_timeout_args: tuple | None = None,
    on_timeout_kwargs: dict | None = None,
) -> Callable:
    """Add a timeout parameter to a function and return it.

    :param seconds: optional time limit in seconds or fractions of a
        second. If None is passed, no timeout is applied.
        This adds some flexibility to the usage: you can disable timing
        out depending on the settings.
    :type seconds: float
    :param on_timeout: optional function to call when the timeout is
        reached instead of raising an exception. If None is passed,
        the default behavior is to raise a TimeoutError exception.
    :type on_timeout: Callable
    :param use_signals: flag indicating whether signals should be used
        for timing function out or the multiprocessing.
        When using multiprocessing, timeout granularity is limited to
        10ths of a second.
    :type use_signals: bool
    :param exception_type: optional exception to raise when the timeout
        is reached. If None is passed, the default behavior is to raise
        a TimeoutError exception.
    :type exception_type: type
    :param exception_message: optional message to pass to the exception
        when the timeout is reached.
    :param on_timeout_args: optional arguments to pass to the on_timeout
        function.
    :type on_timeout_args: tuple
    :param on_timeout_kwargs: optional keyword arguments to pass to the
        on_timeout function.
    :type on_timeout_kwargs: dict

    :return: wrapped function

    :raises: TimeoutError if time limit is reached

    It is illegal to pass anything other than a function as the first
    parameter. The function is wrapped and returned to the caller.
    """
    if not issubclass(exception_type, Exception):
        raise TypeError("exception_type must be a subclass of Exception")

    def decorate(function: Callable) -> Callable:
        if on_timeout is None:
            return _signaler(
                function,
                seconds=seconds,
                use_signals=use_signals,
                exception_type=exception_type,
                exception_message=exception_message,
            )
        return _exception_handler(
            _signaler(
                function,
                seconds=seconds,
                use_signals=use_signals,
                exception_type=exception_type,
                exception_message=exception_message,
            ),
            on_timeout=on_timeout,
            exception_type=exception_type,
            on_timeout_args=on_timeout_args,
            on_timeout_kwargs=on_timeout_kwargs,
        )

    return decorate



"""This module contains the class used to wrap a multithreaded function."""
from __future__ import annotations

import multiprocessing
import sys
import time
from typing import Callable

from timeoutd.exceptions import _raise_exception


def _target(queue, function, *args, **kwargs) -> None:
    """Run a function with arguments and return output via a queue.

    This is a helper function for the Process created in _Timeout. It
    runs the function with positional arguments and keyword arguments
    and then returns the function's output by way of a queue. If an
    exception gets raised, it is returned to _Timeout to be raised by
    the value property.
    """
    try:
        queue.put((True, function(*args, **kwargs)))
    except Exception:  # pylint: disable=broad-except
        queue.put((False, sys.exc_info()[1]))


class _Timeout:  # pylint: disable=too-many-instance-attributes
    """Wrap a function and add a timeout (limit) attribute to it.

    Instances of this class are automatically generated by the
    add_timeout function defined above. Wrapping a function allows
    asynchronous calls to be made and termination of execution after a
    timeout has passed.

    :param function: function to wrap
    :type function: Callable
    :param exception_type: exception to raise when the timeout is
        reached.
    :type exception_type: type
    :param exception_message: optional message to pass to the exception
        when the timeout is reached.
    :type exception_message: str
    :param limit: optional time limit in seconds or fractions of a
        second. If None is passed, no timeout is applied.
    :type limit: float
    """

    def __init__(
        self,
        function: Callable,
        exception_type: type,
        exception_message: str | None,
        limit: float | None,
    ):  # pylint: disable=too-many-arguments
        """Initialize instance in preparation for being called."""
        self.__limit = limit
        self.__function = function
        self.__exception_type = exception_type
        self.__exception_message = exception_message
        self.__name__ = function.__name__
        self.__doc__ = function.__doc__
        self.__timeout = time.time()
        self.__process = multiprocessing.Process()
        self.__queue: multiprocessing.Queue = multiprocessing.Queue()

    def __call__(self, *args, **kwargs):
        """Execute the embedded function object asynchronously.

        The function given to the constructor is transparently called
        and requires that "ready" be intermittently polled. If and when
        it is True, the "value" property may then be checked for
        returned data.
        """
        self.__limit = kwargs.pop("timeout", self.__limit)
        self.__queue = multiprocessing.Queue(1)
        args = (self.__queue, self.__function) + args
        self.__process = multiprocessing.Process(
            target=_target, args=args, kwargs=kwargs
        )
        self.__process.daemon = True
        self.__process.start()
        if self.__limit is not None:
            self.__timeout = self.__limit + time.time()
        while not self.ready:
            time.sleep(0.01)
        return self.value

    def cancel(self):
        """Terminate any possible execution of the embedded function."""
        if self.__process.is_alive():
            self.__process.terminate()

        _raise_exception(self.__exception_type, self.__exception_message)

    @property
    def ready(self):
        """Read-only property indicating status of "value" property."""
        if self.__limit and self.__timeout < time.time():
            self.cancel()
        return self.__queue.full() and not self.__queue.empty()

    @property
    def value(self):
        """Read-only property containing data returned from function."""
        if self.ready is True:
            flag, load = self.__queue.get()
            if flag:
                return load
            raise load
        return None
